open Sharing
open Term

exception RecursiveType

let free_vars t = 
  if Sharing.cyclic t then raise RecursiveType else 
  let rec aux acc t = 
    let r = repr t in 
    match !r with 
      | Repr Var -> r::acc
      | Repr (Op (_, l)) -> 
          List.fold_left aux acc l 
      | Link _ -> assert false
  in aux [] t

let instantiate (l, t) = 
  if Sharing.cyclic t then raise RecursiveType else 
  let l' = List.map (fun x -> (x, Type.new_var ())) l in 
  let rec aux t = 
    let r = repr t in 
    match !r with 
      | Repr Var -> (try List.assq r l' with Not_found -> r)
      | Repr (Op (f, l)) -> ref (Repr (Op (f, List.map aux l)))
      | Link _ -> assert false
  in aux t
              
let rec infer env : Term.t -> Sharing.t = function 
    | TVar x -> 
        (try 
          instantiate (List.assoc x env)
         with Not_found -> 
          failwith (Printf.sprintf "infer: unbound variable '%s'" x))
    | TCst VUnit -> Type.type_unit
    | TCst (VInt _) -> Type.type_int
    | TCst (VBool _) -> Type.type_bool
    | TCst _ -> failwith "infer: cannot type non-base value"
    | TApp (t1,t2) ->
        let tau1 = infer env t1 in 
        let tau2 = infer env t2 in 
        let alpha = Type.new_var () in 
        unify tau1 (Type.build_arrow tau2 alpha);
        alpha
   | TFun (x, t) -> 
        let alpha = Type.new_var () in 
        let env = (x,([], alpha))::env in 
        Type.build_arrow alpha (infer env t)
   | TLet (x, t1, t2) -> 
        let tau1 = infer env t1 in 
        let env = (x, (free_vars tau1, tau1))::env in 
        infer env t2
   | TIf (t1, t2, t3) -> 
        let tau1 = infer env t1 in 
        let tau2 = infer env t2 in 
        let tau3 = infer env t3 in 
        unify tau1 Type.type_bool; 
        unify tau2 tau3;
        tau2
   | TAdd (t1, t2) | TSub (t1, t2) | TMul (t1, t2) -> 
        unify Type.type_int (infer env t1);
        unify Type.type_int (infer env t2);
        Type.type_int
   | TEq (t1, t2) | TLeq (t1, t2) -> 
        unify Type.type_int (infer env t1);
        unify Type.type_int (infer env t2);
        Type.type_bool
   | TLetTuple (x, y, t1, t2) -> 
        let alpha, beta = Type.new_var (), Type.new_var () in 
        let prod = Type.build_product alpha beta in 
        let tau1 = infer env t1 in  
        unify prod tau1;
        let env = (x,(free_vars alpha, alpha))::
                  (y,(free_vars beta, beta))::env in 
        infer env t2
   | TPair (t1, t2) -> 
        Type.build_product (infer env t1) (infer env t2)
   | TFix (f,x,t) -> 
        let alpha = Type.new_var () in 
        let beta = Type.new_var () in 
        let arrow = Type.build_arrow alpha beta in 
        let env = (f,([], arrow))::(x,([], alpha))::env in 
        unify beta (infer env t); arrow
   | TRef t1 -> 
        Type.build_ref (infer env t1)
   | TDeref t1 -> 
        let alpha = Type.new_var () in 
        let refe = Type.build_ref alpha in 
        unify refe (infer env t1);
        alpha
   | TAss (t1, t2) -> 
        let alpha = Type.new_var () in 
        let refe = Type.build_ref alpha in 
        unify refe (infer env t1);
        unify alpha (infer env t2);
        Type.type_unit
   | TSeq (t1, t2) -> 
        ignore (infer env t1); infer env t2




